/*
 * cryptography module for innova,
 *
 * support addictional cryptography on choosed session,
 * use local file as key, options should check evry X time
 * precence of file, for autenticate not only what the
 * used know (the password for login) but also what the user
 * have (the usb file storage with the key file)
 *
 * the options are:
 *
 * timeout XX
 *
 * if present, XX is the number of timeout expired for check 
 * presence of keyfile
 *
 * keyfile
 *
 * is required, and take as argument the file for initialize
 * key for session ecnryption
 *
 * from the file is initialize a buffer key, expanded for 
 * init encrypion algorithm
 */


#include <stdio.h>
#include <string.h>
#include <mcrypt.h>
#include "innova.h"

static char *plug_desc="encryption module";

/* I've taken example from man page of libmcrypt */
static MCRYPT td;
static char *IV;
#define MAXKEYLEN 1024
static unsigned char keybuf[MAXKEYLEN];
static char *cipher_name;
static FILE *keyf;
static int state_size;

#define CIPHER_NUMBER	18
const char *algo_list[CIPHER_NUMBER]= { 
	"des", "tripledes", "cast-128", "cast-256", "xTEA", "blowfish", 
	"twofish", "loki97", "rc2", "rijndael-128", "rijndael-192", "rijndael-256", 
	"serpent", "gost", "safer-sk64", "safer-sk128", "saferplus", "threeway"
	};

char *get_io_desc(int *pv)
{
	*pv =PLUGIN_FORMAT;
	return plug_desc;
}

int mangle_init(struct innova_struct *is)
{
	int i;
	unsigned char foo[64];
	
	innova_p(INFO, "init %s cipher with %d bytes of key", cipher_name, MAXKEYLEN);
	
	if((td= mcrypt_module_open(cipher_name, NULL, "cfb", NULL)) ==MCRYPT_FAILED)
		return -1;

	if(mcrypt_enc_get_iv_size(td) > MAXKEYLEN)
		innova_p(FATAL, "cipher %s require %d byte of key", cipher_name, mcrypt_enc_get_iv_size(td));

	/* innova malloc, simple handle error in the api and fill to 0x00 buffer */
	imalloc((void **)&IV, mcrypt_enc_get_iv_size(td));

	/* 
	 * for generate a repetible random, I use the key, but not the first X byte, but
	 * the first byte since the ascii value of the first key byte, for this reason
	 * MAXKEYLEN must not going under 255 
	 */
	for(i =*keybuf; i < (mcrypt_enc_get_iv_size(td) + *keybuf); i++)
		IV[i - *keybuf] =keybuf[i];

	if((i= mcrypt_generic_init(td, keybuf, mcrypt_enc_get_iv_size(td), IV)) < 0)
		innova_p(FATAL, "%s: mcrypt_generic_init: %s", plug_desc, mcrypt_strerror(i));

	mcrypt_enc_get_state(td, foo, &state_size);
	
	/*
	 * return < 0 is error, and innova is break,
	 * return 0 is for repeat mangle_init, until ret change
	 * return > 0 is for init success, innova continue happy
	 */
	return 1;
}

int mangle_cleanup(struct innova_struct *is, int error)
{
	switch(error)
	{
		case 0x00:
			/* user request causes interrupt */
			break;
		case 0xffff:
			innova_p(ERROR, "program fault, closing...");
			break;
		default:
			innova_p(ERROR_ERRNO, "%s received fatal error", plug_desc);
			break;
	}
	return -1;
}

/*
 * mangle function on crypto plugin, simpy decipher and encipher data...
 * local encrypt and remote decrypt. 
 */

int local_mangle(struct innova_struct *is, struct innova_packet *pkt)
{
	unsigned char *state;

	if(!pkt->data_len)
		return PACKET_OK;

	state =&keybuf[pkt->tcp->seq % (MAXKEYLEN -state_size)];

	mcrypt_enc_set_state(td, state, state_size);

	mcrypt_generic(td, pkt->data, pkt->data_len);

	return PACKET_OK;
}

int remote_mangle(struct innova_struct *is, struct innova_packet *pkt)
{
	unsigned char *state;

	if(!pkt->data_len)
		return PACKET_OK;

	state =&keybuf[pkt->tcp->seq % (MAXKEYLEN -state_size)];
	
	mcrypt_enc_set_state(td, state, state_size);

	mdecrypt_generic(td, pkt->data, pkt->data_len);
	
	return PACKET_OK;
}

int foo_mangle_f(unsigned char hook, struct innova_struct *is, struct innova_packet *pkt, void *x)
{
	/*
	 * possibile hook:
	 *
	 * INIT_HOOK, CLEANUP_HOOK, OPT_PARSE_HOOK, LOCAL_M_HOOK,
	 * REMOTE_M_HOOK, UNKNOW_M_HOOK, TIMEEXC_HOOK
	 */

	return PACKET_OK;
}

/*
 * return value lower than zero mean error, on this case timeout should be
 * used for key recomputation, but we have not a control channel with 
 * the remote host and race condition should cause desyncronization, should
 * be implement on the future. or should be implemented a simple keep alive
 * (this is very easy...)
 */
int io_timeexceed(struct innova_struct *is, int *timeout)
{
	return 0x00;
}

void crypto_plugin_help(void)
{
	int i;

	fprintf(stderr, 
		"%s need some option:\n"
		"\tkey <key_file>\t\trequired.\n"
		"\tcipher <name>\t\t(default Rijndael/AES)\n\n"
		"other cipher should be choosen with libmcrypt(3):\n\t",
	plug_desc);

	/* print with nice indentation :) */
	for(i =0x00; i < CIPHER_NUMBER; i++)
	{
		fprintf(stderr, "%s ", algo_list[i]);
		if(!((i +1)  % (CIPHER_NUMBER / 3)))
			fprintf(stderr, "\n\t");
	}

	fprintf(stderr, "\n");
	fprintf(stderr, "keyfile is how a password, simply, more bigger :) "
			"usually I generate them with /dev/urandom\n\n");
}

int option_parser(struct innova_struct *is, struct innova_options *iopt)
{
	int i;

	if(iopt->plug_opt ==NULL)
	{
		crypto_plugin_help();
		return 0xff;
	}

	for(i =0x00; i < iopt->argc; i++)
	{
		char *optarg, *optind;
		
		optind =iopt->argv[i];
		optarg =iopt->argv[i +1];
		
		if(optind ==NULL || optarg ==NULL)
			break;

		if(!strcasecmp("key", optind))
		{
			keyf =fopen(optarg, "r");

			if(keyf ==NULL)
				innova_p(FATAL_ERRNO, "unable to open key file %s", optarg);

			/* keybuf and MAXKEYLEN is on the top of this file */
			if(fread(keybuf, sizeof(char), MAXKEYLEN, keyf) !=MAXKEYLEN)
				innova_p(FATAL_ERRNO, "key %s must be almost %d byte", optarg, MAXKEYLEN);

			fclose(keyf);
		}

		if(!strcasecmp("cipher", optind))
			cipher_name =optarg;
	}

	if(keyf ==NULL)
	{
		crypto_plugin_help();
		return 0xff;
	}

	cipher_name = cipher_name == NULL ? "rijndael-192" : cipher_name;

	return 0x00;
}
