/*
 * Wed Oct 22 15:39:10 CEST 2003
 *
 * Claudio Agosti <vecna@s0ftpj.org>
 *
 * plugin for change string matched on session stream, 
 */

#include <stdio.h>
#include <unistd.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <netinet/ip.h>
#include <netinet/tcp.h>

#include "innova.h"

/* string matcher plugin */
static char *plug_desc="string matcher";

/* info about pattern tracking and string substitution */
struct couple
{
#define BIDIR_CHECK	0
#define INPUT_CHECK	1
#define OUTPUT_CHECK	2
	unsigned char IO;
	char *tomatch;
	char *toput;
	unsigned int time;
};
#define MAXCOUPLELIST	32
#define MAXLINE		4096

static struct couple pattern_info[MAXCOUPLELIST];
static unsigned short patter_number;
static char *pattern_file;

/* 
 * info about connection tracking, all session hijacked must be keep
 * tracked because one time the string is changed out kernel keep
 * desynched sequence number
 *
 */

char *get_io_desc(int *pv)
{
	*pv =PLUGIN_FORMAT;
	return plug_desc;
}

int mangle_init(struct innova_struct *is)
{
	FILE *datafile;
	int line_number =0;

	printf(" %s innova plugins is reading patterns\n", plug_desc);

	if((datafile =fopen(pattern_file, "r")) ==NULL)
	{
		printf("error on loading pattern file %s\n", pattern_file);
		return -1;
	}

	while(1)
	{
		char *p, backup, line[MAXLINE];
		int i, first, second check =0;

		fgets(line, MAXLINE, datafile);

		if(line[0] =='#' || line[0] =='\n')
			continue;

		line_number++;

		for(i =0; i < strlen(line); i++)
		{
			/* very BAD routing for read pattern file, but work :) */
			if(line[i] =='\"' && (!i || line[i -1] !='\\'))
			{
				check++;

				if(!first)
					first =i;
				else
				{
					backup =line[i];

					line[i] =0x00;

					switch(check)
					{
						case 2:
							pattern_info[pattern_number].tomatch =strdup(&line[first]);
							break;
						case 4:
							pattern_info[pattern_number].toput =strdup(&line[first]);
							break;
						deafult:
							printf("bad patter specification on line %d\n", line_number);
							return -1;
					}

					first =0;
					line[i] =backup;
				}
			}

			/* "to match" "to put" input|output|both NUMBER */
			if(check == 4 && line[i] ==' ')
			{
				if(!strcmp(&line[i], " input"))
				{
					i +=strlen(" input ");
					pattern_info[pattern_number].IO =INPUT_CHECK;
				}
				else if(!strcmp(&line[i], " output"))
				{
					i +=strlen(" output ");
					pattern_info[pattern_number].IO =OUTPUT_CHECK;
				}
				else if(!strcmp(&line[i], " bidir"))
				{
					i +=strlen(" bidir ");
					pattern_info[pattern_number].IO =BIDIR_CHECK;
				}
				else
				{
					printf("invalid direction specification (input|output|bidir) at line %d\n", 
						line_number);
					return -1;
				}

				/* 0 time specifie involv infinite matching/subsistution */
				patern_info[pattern_number].time =atoi(&line[i]);
			}
		}

		if(MAXCOUPLELIST == (pattern_number -1))
		{
			printf("your pattern file have a lot of pattern, actual max is set to %d\n"
		               "to avoid mistake, I quit. try to comment superfluos line, or "
			       "change MAXCOUPLELIST on %s source file.\n",
			       MAXCOUPLELIST, __FILE__
			);
			return -1;
		}

		if(check != 4)
		{
			printf("line %d ignored (bad pattern file format)\n", line_number);
			continue;
		}

		if(feof(datafile))
			break;
	}

	fclose(datafile);
	/*
	 * return < 0 is error, and innova is break,
	 * return 0 is for repeat mangle_init,
	 * return > 0 is for init success, innova continue happy
	 */
	return 1;
}

int mangle_cleanup(struct innova_struct *is, int error)
{
	memset(pattern_info, 0x00, sizeof(pattern_info));

	pattern_number =0;

	return -1;
}

int local_mangle(struct innova_struct *is, struct innova_packet *pkt)
{
	return PACKET_OK;
}

int remote_mangle(struct innova_struct *is, struct innova_packet *pkt)
{
	return PACKET_OK;
}

/*
 * foo_mangle is function called before any other function, I'm using this because 
 * on the remote_mangle and on the local_mangle I must check for duplicated packets
 * (because duplicated packet mean a lost of previous packet for remote peer, and
 * retransission of steganographic segment, and already dumped data, ignored for the
 * receiver)
 */
int foo_mangle_f(unsigned char hook, struct innova_struct *is, struct innova_packet *pkt, void *x)
{
	switch(hook)
	{
		case LOCAL_M_HOOK:
			break;
		case REMOTE_M_HOOK:
			break;
		default:
			return 0x00;
	}

	return PACKET_OK;
}

int io_timeexceed(struct innova_struct *is, int *timeout)
{
	return 0x00;
}

void print_help(void)
{
	fprintf(stderr, "file %s required only one options on innova command line: "
			"the file with patter to match\n"
			"this plugin file want be and example for string matching/string injection"
			" and connection tracking\n"
			"on innova plugin, for work is required one configuration file with this format:\n"
			"\"pattern to match\" \"pattern to put\" DIRECTIONAL_KEYWORD NUMBER_OF_TIME\n\n"
			"DIRECTIONAL_KEYWORD should be \"input\" \"output\" \"bidir\" and is required for\n"
			"specify on witch flow of session innova must check and inject.\n"
			"NUMBER_OF_TIME is the number of time who innova must exchange the pattern on each session\n"
			"if you put 0, is mean infinite time.\n\n", __FILE__
	);
}

int option_parser(struct innova_struct *is, struct innova_options *iopt)
{
	int ret;

	if(iopt->plug_opt ==NULL || !strcmp(iopt->plug_opt, "help"))
	{
		print_help();
		ret =0xff;
	}
	else
	{
		pattern_file =strdup(iopt->plug_opt);
		ret =0x00;
	}

	return ret;
}
